import logging
from json.decoder import JSONDecodeError
from typing import List, Dict

import requests
from outcache import Cache

from .models import Player, Score
from .common import Common


class ScoreSaber:
    TIMEOUT = 10
    _url = "https://new.scoresaber.com/api"

    def __init__(self):
        self.log = logging.getLogger(__name__)

    def _process_url(self, url: str) -> Dict:
        response = Common.request(requests.get, url, timeout=self.TIMEOUT)

        try:
            data = response.json()
        except JSONDecodeError:
            self.log.exception("JSONDecodeError, response: %r, response.text: %r", response, response.text)
            data = {"error": "Failed to decode json from scoresaber. Somethings broken."}

        return data

    @Cache(minutes=2)
    def _get_player_basic(self, player_id: str) -> Dict:
        return self._process_url(f"{self._url}/player/{player_id}/basic")

    def get_player_basic(self, player_id: str) -> Player:
        response = self._get_player_basic(player_id)

        return Player.from_dict(response["playerInfo"])

    @Cache(minutes=2)
    def _get_player_full(self, player_id: str) -> Dict:
        return self._process_url(f"{self._url}/player/{player_id}/full")

    def get_player_full(self, player_id: str) -> Player:
        response = self._get_player_full(player_id)

        return Player.from_dict(response["playerInfo"])

    @Cache(minutes=2)
    def _get_recent_scores(self, player_id: str, page: int = 1) -> Dict:
        return self._process_url(f"{self._url}/player/{player_id}/scores/recent/{page}")

    def get_recent_scores(self, player_id: str, page: int = 1) -> List[Score]:
        response = self._get_recent_scores(player_id, page)

        recent_score_list = []

        for recent_score in response["scores"]:
            recent_score_list.append(Score.from_dict(recent_score))

        return recent_score_list

    @Cache(minutes=2)
    def _get_top_scores(self, player_id, page: int = 1) -> Dict:
        return self._process_url(f"{self._url}/player/{player_id}/scores/top/{page}")

    def get_top_scores(self, player_id, page: int = 1) -> List[Score]:
        response = self._get_top_scores(player_id, page)

        top_score_list = []

        for top_score in response["scores"]:
            top_score_list.append(Score.from_dict(top_score))

        return top_score_list



class APIException(Exception):
    def __init__(self, response, message=None):
        self.response = response

        if message is not None:
            self.message = message
        else:
            self.message = f"Got HTTP status code {response.status_code} for {response.url}"

        super().__init__(self.message)

    def __str__(self):
        return self.message


class NotFoundException(APIException):
    pass


class RateLimitedException(APIException):
    pass


class ServerErrorException(APIException):
    pass



import logging
import time

from .errors import RateLimitedException, NotFoundException, ServerErrorException

WAIT_RATE_LIMIT = 60
WAIT_SERVER_ERROR = 5


class Common:

    @staticmethod
    def request(method, *args, **kwargs):
        def attempt():
            response = method(*args, **kwargs)
            Common.verify_response(response)
            return response

        return Common.request_attempt(attempt)

    @staticmethod
    def verify_response(response):
        if response.status_code == 429:
            raise RateLimitedException(response)

        if 400 <= response.status_code < 500:
            raise NotFoundException(response)

        if 500 <= response.status_code < 600:
            raise ServerErrorException(response)

    @staticmethod
    def request_attempt(func):
        attempt = 0
        attempting = True
        last_exception = None

        while attempting and attempt < 6:
            try:
                result = func()

                attempting = False
                return result
            except ServerErrorException as error:
                attempt += 1
                logging.info(str(error))
                logging.info(f"Waiting {WAIT_SERVER_ERROR} seconds...")
                last_exception = error
                time.sleep(WAIT_SERVER_ERROR)
            except RateLimitedException as error:
                attempt += 1
                logging.info(str(error))
                logging.info(f"Waiting {WAIT_RATE_LIMIT} seconds...")
                last_exception = error
                time.sleep(WAIT_RATE_LIMIT)

        if last_exception is not None:
            raise last_exception



from dateutil import parser
from dataclasses import field, Field
from datetime import datetime
from typing import Optional

from dataclasses_json import config
from marshmallow import fields


def datetime_from_iso_format(time):
    if time:
        return parser.isoparse(time)

    return None


def datetime_field(json_field_name: Optional[str] = None) -> Field:
    if json_field_name is None:
        conf = config(
                encoder=datetime.isoformat,
                decoder=datetime_from_iso_format,
                mm_field=fields.DateTime(format='iso')
            )
    else:
        conf = config(
            encoder=datetime.isoformat,
            decoder=datetime_from_iso_format,
            mm_field=fields.DateTime(format='iso'),
            field_name=json_field_name,
        )

    return field(
        default=None,
        metadata=conf
    )


def default(json_field_name: Optional[str] = None) -> Field:
    if json_field_name is None:
        conf = config()
    else:
        conf = config(field_name=json_field_name)

    return field(
        default=None,
        metadata=conf
    )
